from enum import Enum
import hashlib
import io
import json
import numbers

from typing import List, Optional, Dict
import zipfile

from decentriq_dcr_compiler import (
    compiler,
    LookalikeMediaDataRoom,
    LookalikeMediaResponse,
    LookalikeMediaRequest,
)

from ..client import Client
from ..session import LATEST_GCG_PROTOCOL_VERSION, Session
from ..storage import Key
from ..keychain import Keychain, KeychainEntry
from ..types import (
    CreateMediaComputeJobInput,
    EnclaveSpecification,
    JobId,
    OverlapInsightsCacheKey,
    PublishedDataset,
)
from ..proto import (
    DataRoom,
    parse_length_delimited,
    serialize_length_delimited,
    CreateDcrKind,
    UserAuth,
    GcgRequest,
)
from ..helpers import (
    get_latest_enclave_specs_as_dictionary,
)

__docformat__ = "restructuredtext"


__all__ = [
    "LookalikeMediaDcr",
]


class DatasetType(Enum):
    MATCHING = "matching"
    SEGMENTS = "segments"
    DEMOGRAPHICS = "demographics"
    EMBEDDINGS = "embeddings"
    AUDIENCES = "audiences"


class ExistingLookalikeMediaDcr:
    def __init__(self, id: str, driver_spec: EnclaveSpecification) -> None:
        self.id = id
        self.driver_enclave_spec = {"decentriq.driver": driver_spec}


class LookalikeMediaDcr:
    def __init__(
        self,
        client: Client,
        high_level_representation: str,
        existing_lookalike_media_dcr: ExistingLookalikeMediaDcr = None,
    ) -> None:
        self.client = client
        if existing_lookalike_media_dcr:
            self.auth, _ = client.create_auth_using_decentriq_pki(
                existing_lookalike_media_dcr.driver_enclave_spec
            )
            self.session = client.create_session(
                self.auth, existing_lookalike_media_dcr.driver_enclave_spec
            )
            self.hl_lmdcr = LookalikeMediaDataRoom.parse_obj(high_level_representation)
            self.id = existing_lookalike_media_dcr.id
        else:
            enclave_specs = get_latest_enclave_specs_as_dictionary(self.client)
            self.auth, _ = client.create_auth_using_decentriq_pki(enclave_specs)
            self.session = client.create_session(self.auth, enclave_specs)
            (self.hl_lmdcr, self.id) = self._create_lmdcr(high_level_representation)
        self.overlap_insights_job = None

    @staticmethod
    def send_request(
        lmdcr_request: LookalikeMediaRequest,
        session: Session,
    ) -> LookalikeMediaResponse:
        pki = session._get_message_pki(session.auth)
        endorsements = session.auth.endorsements
        auth = UserAuth(pki=pki, enclaveEndorsements=endorsements)
        request_serialized = compiler.compile_lookalike_media_request_serialized(
            lmdcr_request,
            serialize_length_delimited(auth),
        )
        gcg_request = GcgRequest()
        parse_length_delimited(request_serialized, gcg_request)
        responses = session.send_request(gcg_request, LATEST_GCG_PROTOCOL_VERSION)
        if len(responses) != 1:
            raise Exception("Malformed response")
        serialised_response = serialize_length_delimited(responses[0])
        response = compiler.decompile_lookalike_media_response(
            lmdcr_request, serialised_response
        )
        return response

    def _create_lmdcr(
        self, high_level_representation: str
    ) -> (LookalikeMediaDataRoom, str):
        create_lmdcr = compiler.CreateLookalikeMediaDataRoom.parse_obj(
            high_level_representation
        )
        lmdcr = compiler.create_lookalike_media_data_room(create_lmdcr)
        lmdcr_serialised = compiler.compile_lookalike_media_data_room(lmdcr)
        low_level_dcr = DataRoom()
        parse_length_delimited(lmdcr_serialised, low_level_dcr)
        lmdcr_id = self.session.publish_data_room(
            low_level_dcr,
            kind=CreateDcrKind.LOOKALIKE_MEDIA,
            high_level_representation=lmdcr.json().encode(),
        )
        return (lmdcr, lmdcr_id)

    def provision_from_data_lab(self, data_lab_id: str, keychain: Keychain):
        """
        Provision the DataLab with the given ID to the Lookalike Media DCR.

        **Parameters**:
        - `data_lab_id`: ID of the DataLab to provision to the Lookalike Media DCR.
        - `keychain`: Keychain to use to provision datasets from the DataLab.
        """
        # Check DataLab is validated
        data_lab = self.client.get_data_lab(data_lab_id)
        if not data_lab["isValidated"]:
            raise Exception("Cannot provision DataLab, not validated.")

        # Check compatibility
        hl_data_lab = data_lab["highLevelRepresentationAsString"]
        compatible = (
            compiler.is_data_lab_compatible_with_lookalike_media_data_room_serialized(
                hl_data_lab, self.hl_lmdcr.json()
            )
        )
        if not compatible:
            raise Exception("DataLab is incompatible with Lookalike Media DCR")

        # Provision datasets
        lmdcr_datasets = compiler.get_consumed_datasets(self.hl_lmdcr.json())
        data_lab_datasets = self._get_data_lab_datasets_dict(data_lab)

        # Check all required datasets can be provisioned by the DataLab before
        # actually provisioning. This is necessary because we could otherwise
        # provision only some of the datasets resulting in a "broken" LMDCR.
        for required_dataset in lmdcr_datasets.required:
            if not data_lab_datasets[required_dataset]["manifestHash"]:
                # The manifest hash of the dataset could not be found in the database.
                raise Exception(
                    f"Unable to provision the required dataset '{required_dataset}' from the DataLab."
                )

        # Provision all required datasets.
        for required_dataset in lmdcr_datasets.required:
            lmdcr_node_name = self._get_lmdcr_node_name(required_dataset)
            manifest_hash = data_lab_datasets[required_dataset]["manifestHash"]
            retrieved_key = keychain.get("dataset_key", manifest_hash)
            self.session.publish_dataset(
                self.id, manifest_hash, lmdcr_node_name, Key(retrieved_key.value)
            )

        # Provision optional datasets if the DataLab is able to.
        for optional_dataset in lmdcr_datasets.optional:
            if (
                optional_dataset not in data_lab_datasets
                or not data_lab_datasets[optional_dataset]
            ):
                # If we can't provision the dataset, move on to the next.
                continue
            lmdcr_node_name = self._get_lmdcr_node_name(optional_dataset)
            manifest_hash = data_lab_datasets[optional_dataset]["manifestHash"]
            retrieved_key = keychain.get("dataset_key", manifest_hash)
            self.session.publish_dataset(
                self.id, manifest_hash, lmdcr_node_name, Key(retrieved_key.value)
            )

        # Update DB.
        provisioned_data_lab_id = self.client._provision_data_lab(self.id, data_lab_id)[
            "id"
        ]
        if provisioned_data_lab_id != data_lab_id:
            raise Exception(
                f"Incorrectly provisioned DataLab {provisioned_data_lab_id}"
            )

    def deprovision_data_lab(self):
        """
        Deprovision a DataLab from the Lookalike Media DCR.
        """
        lmdcr_datasets = compiler.get_consumed_datasets(self.hl_lmdcr.json())
        # Deprovision all required datasets.
        for required_dataset in lmdcr_datasets.required:
            lmdcr_node_name = self._get_lmdcr_node_name(required_dataset)
            self.session.remove_published_dataset(self.id, lmdcr_node_name)

        # Deprovision optional datasets.
        for optional_dataset in lmdcr_datasets.optional:
            lmdcr_node_name = self._get_lmdcr_node_name(optional_dataset)
            self.session.remove_published_dataset(self.id, lmdcr_node_name)

        # Update DB.
        data_lab = self.client._deprovision_data_lab(self.id)
        if data_lab:
            raise Exception("DataLab should have been deprovisioned")

    def _get_data_lab_datasets_dict(self, data_lab: compiler.DataLab):
        datasets_dict = {}
        for dataset in data_lab["datasets"]:
            datasets_dict[dataset["name"]] = dataset["dataset"]
        return datasets_dict

    def _get_lmdcr_node_name(self, dataset_name: str):
        lmdcr_node_name = (
            compiler.get_lookalike_media_node_names_from_data_lab_data_type(
                dataset_name
            )
        )
        if not lmdcr_node_name:
            raise Exception(f"Unknown LMDCR node name for dataset name {dataset_name}")
        return lmdcr_node_name

    def generate_insights(self):
        """
        Run the overlap insights computation.
        Use the `get_overlap_insights` method to retrieve the results of this computation.
        """
        if not self._has_insights_generation_support():
            raise Exception("LMDCR does not support insights generation")

        scope_id = self.client._ensure_dcr_data_scope(self.id)
        lmdcr_request = LookalikeMediaRequest.parse_obj(
            {
                "calculateOverlapInsights": {
                    "dataRoomIdHex": self.id,
                    "scopeIdHex": scope_id,
                }
            }
        )
        response = LookalikeMediaDcr.send_request(lmdcr_request, self.session)
        response_json = json.loads(response.json())
        job_id_hex = response_json["calculateOverlapInsights"]["jobIdHex"]
        cache_key = self._get_overlap_insights_cache_key_string()
        input = CreateMediaComputeJobInput(
            publishedDataRoomId=self.id,
            computeNodeName="consentless_overlap_insights",
            cacheKey=cache_key,
            jobType="LOOKALIKE_MEDIA_OVERLAP_INSIGHTS",
            jobIdHex=job_id_hex,
        )
        self.overlap_insights_job = self.client._create_media_compute_job(input)

    def _has_insights_generation_support(self) -> bool:
        features = self._get_features()
        return "GENERATE_INSIGHTS_COMPUTATION" in features

    def _get_features(self):
        features = compiler.get_lookalike_media_data_room_features(self.hl_lmdcr)
        return features

    def _get_overlap_insights_cache_key_string(self) -> str:
        published_datasets = self.session.retrieve_published_datasets(
            self.id
        ).publishedDatasets
        cache_key = self._generate_overlap_insights_cache_key(published_datasets)
        return self._compute_cache_key_string(cache_key)

    def _generate_overlap_insights_cache_key(
        self, published_datasets
    ) -> OverlapInsightsCacheKey:
        cache_key = OverlapInsightsCacheKey()
        cache_key["dataRoomId"] = self.id
        published_datasets_list: List[PublishedDataset] = []
        for dataset in published_datasets:
            published_dataset = PublishedDataset()
            published_dataset["leafId"] = dataset.leafId
            published_dataset["user"] = dataset.user
            published_dataset["timestamp"] = dataset.timestamp
            published_dataset["datasetHash"] = bytearray(dataset.datasetHash)
            published_datasets_list.append(published_dataset)

            node_id = dataset.leafId
            dataset_hash = dataset.datasetHash.hex()
            if node_id == "audiences":
                cache_key["advertiserDatasetHash"] = dataset_hash
            elif node_id == "matching":
                cache_key["publisherUsersDatasetHash"] = dataset_hash
            elif node_id == "segments":
                cache_key["publisherSegmentsDatasetHash"] = dataset_hash
            elif node_id == "embeddings":
                cache_key["publisherEmbeddingsDatasetHash"] = dataset_hash
            elif node_id == "demographics":
                cache_key["publisherDemographicsDatasetHash"] = dataset_hash
        cache_key["publishedDatasets"] = published_datasets_list
        return cache_key

    # TODO: Add this function to DDC so it is common between the JS and Python clients.
    #       At the moment the functionality is implemented twice, which could lead to a mismatch
    #       in behaviour.
    def _compute_cache_key_string(self, key) -> str:
        def cache_key_hashed(digest, key):
            if isinstance(key, str):
                digest.update(bytes(key, "UTF-8"))
                return
            elif isinstance(key, bool):
                digest.update(str(key))
                return
            elif isinstance(key, numbers.Number):
                digest.update(bytes(str(key), "UTF-8"))
                return
            elif isinstance(key, list):
                for element in key:
                    cache_key_hashed(digest, element)
                # Do it a second time as that is what the JS code does.
                # https://github.com/decentriq/delta/blob/c9d4c56703ccbbe9c0a2bf65daba18036d9c7aac/avato-backend/frontend/decentriq-platform/src/wrappers/ApolloWrapper/resolvers/LruCache.ts#L19-L26
                for element in key:
                    cache_key_hashed(digest, element)
                cache_key_hashed(digest, len(key))
            elif isinstance(key, bytearray):
                # Sort an index of `str` values so that we get a list similar to:
                # ['0', '1', '10', '11', '12', '13', '14', '15', '2', '3', '4', '5', '6', '7', '8', '9']
                # This sorted index list will be used to determine the order of fields to hash.
                # ** Note: The list is not numerically ordered. **
                index_list = [str(x) for x in list(range(0, len(key)))]
                sorted_index_list = [int(x) for x in sorted(index_list)]
                for index in sorted_index_list:
                    cache_key_hashed(digest, key[index])
            elif isinstance(key, dict):
                for field in dict(sorted(key.items())):
                    cache_key_hashed(digest, key[field])
                return
            elif isinstance(key, bytes):
                digest.update(key)
            else:
                raise Exception(f"Unexpected nested key type {type(key)}, key {key}")

        if isinstance(key, str):
            return key
        digest = hashlib.sha256()
        cache_key_hashed(digest, key)
        hex_digest = digest.hexdigest()
        return hex_digest

    def get_overlap_insights(self, timeout: int = None) -> Dict[str, str]:
        """
        Retrieve the results of running the overlap insights computation.
        """
        if not self._has_insights_generation_support():
            raise Exception("LMDCR does not support insights generation")

        input = {
            "publishedDataRoomId": self.overlap_insights_job["publishedDataRoomId"],
            "jobType": self.overlap_insights_job["jobType"],
            "cacheKey": self.overlap_insights_job["cacheKey"],
        }
        media_compute_job = self.client._get_media_compute_job(input)
        job_id = JobId(media_compute_job["jobIdHex"], "consentless_overlap_insights")
        results = self.session.get_computation_result(
            job_id, interval=5, timeout=timeout
        )
        zip = zipfile.ZipFile(io.BytesIO(results), "r")
        if "overlap.json" in zip.namelist():
            overlap_insights = zip.read("overlap.json").decode()
            report = json.loads(overlap_insights)
            return report
        else:
            raise Exception("Failed to retrieve overlap insights")

    def _get_lmdcr_gcg_request(
        self, lmdcr_request: LookalikeMediaRequest
    ) -> GcgRequest:
        pki = self.session._get_message_pki(self.auth)
        endorsements = self.auth.endorsements
        auth = UserAuth(pki=pki, enclaveEndorsements=endorsements)
        request_serialized = compiler.compile_lookalike_media_request_serialized(
            lmdcr_request,
            serialize_length_delimited(auth),
        )
        request = GcgRequest()
        parse_length_delimited(request_serialized, request)
        return request

    def _upload_dataset_to_keychain(
        self, file_path: str, name: str, key: Key, keychain: Keychain
    ):
        with open(file_path, "rb") as file:
            dataset_id = self.client.upload_dataset(file, key, name)
            keychain.insert(KeychainEntry("dataset_key", dataset_id, key.material))
            return dataset_id

    def retrieve_audit_log(self) -> str:
        """
        Retrieve the Lookalike Media DCR audit log.
        """
        return self.session.retrieve_audit_log(self.id).log.decode("utf-8")


# Allow provisioning a dataset directly to a Lookalike Media DCR.
# This is not intended for end users. They should use the `provision_from_data_lab` method instead.
def provision_dataset(
    data: io.BytesIO,
    *,
    name: str,
    session: Session,
    key: Key,
    data_room_id: str,
    dataset_type: DatasetType,
    store_in_keychain: Optional[Keychain] = None,
    description: str = "",
) -> str:
    manifest_hash = session.client.upload_dataset(
        data, key, name, description=description, store_in_keychain=store_in_keychain
    )
    session.publish_dataset(
        data_room_id, manifest_hash, leaf_id=dataset_type.value, key=key
    )

    return manifest_hash



from __future__ import annotations

from .high_level_node import ContainerComputationNode
from .script import FileContent, ScriptingLanguage
from ..session import Session
from typing import Optional, List, Dict
from .node_definitions import NodeDefinition
from typing_extensions import Self
from decentriq_dcr_compiler.schemas.data_science_data_room import (
    ScriptingComputationNode,
)
import json


class RComputeNodeDefinition(NodeDefinition):
    def __init__(
        self,
        name: str,
        script: str,
        additional_files: Optional[List[FileContent]] = None,
        dependencies: List[str] = [],
        enable_logs_on_error: bool = False,
        enable_logs_on_success: bool = False,
        output: Optional[str] = "/output",
        id: Optional[str] = None,
    ) -> None:
        """
        Initialise a `RComputeNodeDefinition`:

        **Parameters**:
        - `name`: Name of the `RComputeNodeDefinition`.
        - `script`: The R computation as a string.
        - `additional_files`: Other files that can be used by the `RComputeNodeDefinition`.
        - `dependencies`: Nodes that the `RComputeNodeDefinition` depends on.
        - `enable_logs_on_error`: Enable logs in the event of an error.
        - `enable_logs_on_success`: Enable logs when the computation is successful.
        - `output`: Directory where the results should be written.
        """
        super().__init__(name=name, id=id or name)
        self.script = script
        self.dependencies = dependencies
        self.additional_files = additional_files
        self.enable_logs_on_error = enable_logs_on_error
        self.enable_logs_on_success = enable_logs_on_success
        self.output = output
        self.scripting_specification_id = "decentriq.r-ml-worker-32-32"
        self.static_content_specification_id = "decentriq.driver"

    @property
    def required_workers(self):
        return [self.scripting_specification_id, self.static_content_specification_id]

    def _get_high_level_representation(self) -> Dict[str, str]:
        """
        Retrieve the high level representation of the `RComputeNodeDefinition`.
        """
        additional_files = []
        if self.additional_files:
            additional_files = [
                {
                    "name": file.name,
                    "content": file.content,
                }
                for file in self.additional_files
            ]
        computation_node = {
            "id": self.id,
            "name": self.name,
            "kind": {
                "computation": {
                    "kind": {
                        "scripting": {
                            "additionalScripts": additional_files,
                            "dependencies": self.dependencies,
                            "enableLogsOnError": self.enable_logs_on_error,
                            "enableLogsOnSuccess": self.enable_logs_on_success,
                            "mainScript": {
                                "name": "r-script",
                                "content": self.script,
                            },
                            "output": self.output,
                            "scriptingLanguage": ScriptingLanguage.r.value,
                            "scriptingSpecificationId": self.scripting_specification_id,
                            "staticContentSpecificationId": self.static_content_specification_id,
                        }
                    }
                },
            },
        }
        return computation_node

    @staticmethod
    def _from_high_level(
        id: str,
        name: str,
        node: ScriptingComputationNode,
    ) -> Self:
        """
        Instantiate a `RComputeNodeDefinition` from its high level representation.

        **Parameters**:
        - `name`: Name of the `RComputeNodeDefinition`.
        - `node`: Pydantic model of the `RComputeNode`.
        """
        scripting_node = json.loads(node.model_dump_json())
        return RComputeNodeDefinition(
            id=id,
            name=name,
            script=scripting_node["mainScript"]["content"],
            dependencies=scripting_node["dependencies"],
            additional_files=[
                FileContent(name=script["name"], content=script["content"])
                for script in scripting_node["additionalScripts"]
            ],
            enable_logs_on_error=scripting_node["enableLogsOnError"],
            enable_logs_on_success=scripting_node["enableLogsOnSuccess"],
            output=scripting_node["output"],
        )

    def build(
        self,
        dcr_id: str,
        node_definition: NodeDefinition,
        client: Client,
        session: Session,
    ) -> RComputeNode:
        """
        Construct a RComputeNode from the Node Definition.

        **Parameters**:
        - `dcr_id`: ID of the DCR the node is a member of.
        - `node_definition`: Definition of the R Compute Node.
        - `client`: A `Client` object which can be used to perform operations such as uploading data
            and retrieving computation results.
        - `session`: The session with which to communicate with the enclave.
        """
        return RComputeNode(
            name=self.name,
            dcr_id=dcr_id,
            script=self.script,
            client=client,
            session=session,
            node_definition=node_definition,
            dependencies=self.dependencies,
            additional_files=self.additional_files,
            enable_logs_on_error=self.enable_logs_on_error,
            enable_logs_on_success=self.enable_logs_on_success,
            output=self.output,
            id=self.id,
        )


class RComputeNode(ContainerComputationNode):
    """
    Class representing an R Computation node.

    An RComputeNode is able to run arbitrary R code.
    """

    def __init__(
        self,
        id: str,
        name: str,
        dcr_id: str,
        script: str,
        client: Client,
        session: Session,
        node_definition: NodeDefinition,
        additional_files: Optional[List[FileContent]] = None,
        dependencies: List[str] = [],
        enable_logs_on_error: bool = False,
        enable_logs_on_success: bool = False,
        output: Optional[str] = "/output",
    ) -> None:
        """
        Initialise a `RComputeNode`:

        **Parameters**:
        - `id`: ID of the `RComputeNode`.
        - `name`: Name of the `RComputeNode`.
        - `dcr_id`: ID of the DCR the node is a member of.
        - `script`: The R computation.
        - `client`: A `Client` object which can be used to perform operations such as uploading data
            and retrieving computation results.
        - `session`: The session with which to communicate with the enclave.
        - `node_definition`: Definition of the Matching Node.
        - `additional_files`: Other files that can be used by the `RComputeNode`.
        - `dependencies`: Nodes that the `RComputeNode` depends on.
        - `enable_logs_on_error`: Enable logs in the event of an error.
        - `enable_logs_on_success`: Enable logs when the computation is successful.
        - `output`: Directory where the results should be written.
        """
        super().__init__(
            name=name,
            client=client,
            session=session,
            dcr_id=dcr_id,
            id=id,
            definition=node_definition,
        )
        self.script = script
        self.dependencies = dependencies
        self.additional_files = additional_files
        self.enable_logs_on_error = enable_logs_on_error
        self.enable_logs_on_success = enable_logs_on_success
        self.output = output
        self.scripting_specification_id = "decentriq.r-ml-worker-32-32"
        self.static_content_specification_id = "decentriq.driver"

    def _get_computation_id(self) -> str:
        """
        Retrieve the ID of the node corresponding to the computation.
        """
        return f"{self.id}_container"



from google.protobuf.json_format import MessageToDict
from ...proto import (
    serialize_length_delimited,
    ComputeNodeFormat,
    parse_length_delimited,
)
from ..node import Node
from .proto import (
    SalesforceWorkerConfiguration,
    ImportRole,
    Core,
    MarketingCloud,
)
from enum import Enum


class ProductType(Enum):
    CORE = 1
    MARKETING_CLOUD = 2


class DataSourceSalesforce(Node):
    """
    Compute node that fetches a dataset from Salesforce.
    """

    def __init__(
        self,
        name: str,
        credentialsDependency: str,
        domainUrl: str,
        apiName: str,
        productType: ProductType,
    ) -> None:
        importRole = (
            ImportRole(core=Core())
            if productType == ProductType.CORE
            else ImportRole(marketingCloud=MarketingCloud())
        )
        config = SalesforceWorkerConfiguration(
            credentialsDependency=credentialsDependency,
            importRole=importRole,
            domainUrl=domainUrl,
            apiName=apiName,
        )
        config_serialized = serialize_length_delimited(config)
        super().__init__(
            name,
            config=config_serialized,
            enclave_type="decentriq.salesforce-worker",
            dependencies=[credentialsDependency],
            output_format=ComputeNodeFormat.RAW,
        )


class DataSourceSalesforceWorkerDecoder:
    def decode(self, config: bytes):
        config_decoded = SalesforceWorkerConfiguration()
        parse_length_delimited(config, config_decoded)
        return MessageToDict(config_decoded)


__all__ = [
    "DataSourceSalesforce",
]



from .builder import LookalikeMediaDcrBuilder
from .lookalike_media import LookalikeMediaDcr, provision_dataset, DatasetType


__pdoc__ = {
    "builder": False,
    "lookalike_media": False,
}

__all__ = [
    "LookalikeMediaDcr",
    "LookalikeMediaDcrBuilder",
    "provision_dataset",
    "DatasetType",
]



import requests
from enum import Enum
from urllib3.util import Retry
from .config import (
    DECENTRIQ_REQUEST_RETRY_TOTAL,
    DECENTRIQ_REQUEST_RETRY_BACKOFF_FACTOR,
)

retry = Retry(
    total=DECENTRIQ_REQUEST_RETRY_TOTAL,
    backoff_factor=DECENTRIQ_REQUEST_RETRY_BACKOFF_FACTOR,
)


class Endpoints(str, Enum):
    GRAPHQL = "/graphql"
    SESSION_MESSAGES = "/sessions/:sessionId/messages"
    USER_UPLOAD_CHUNKS = "/uploads/:uploadId/chunks/:chunkHash"


class ApiError(Exception):
    pass


class AuthorizationError(ApiError):
    pass


class NotFoundError(ApiError):
    pass


class BadRequestError(ApiError):
    pass


class ServerError(ApiError):
    pass


class Api:
    def __init__(
        self,
        api_token,
        client_id,
        host,
        port,
        use_tls,
        api_prefix,
        additional_auth_headers={},
        timeout=None,
    ):
        session = requests.Session()
        if use_tls:
            protocol = "https"
        else:
            protocol = "http"
        self.base_url = f"{protocol}://{host}:{port}{api_prefix}"
        auth_headers = {
            "Authorization": "Bearer " + api_token,
            "Authorization-Type": "app",
            "Authorization-Client": client_id,
        }
        auth_headers.update(additional_auth_headers)
        session.headers.update(auth_headers)
        self.session = session
        self.timeout = timeout

    @staticmethod
    def __check_response_status_code(response):
        body = response.content
        try:
            payload = response.json()
            if "errors" in payload:
                errors = payload["errors"]
                message = ",".join([error["message"] for error in errors])
                body = message
        except:
            pass

        if response.status_code >= 200 and response.status_code <= 204:
            pass
        elif response.status_code == 400:
            raise BadRequestError(body)
        elif response.status_code == 401 or response.status_code == 403:
            raise AuthorizationError(body)
        elif response.status_code == 404:
            raise NotFoundError(body)
        elif response.status_code >= 500 and response.status_code <= 504:
            raise ServerError(body)
        else:
            raise ApiError(body)

    def _request(self, method, endpoint, **kwargs):
        retry = kwargs.pop("retry", None)

        url = self.base_url + endpoint
        try:
            response = self.session.request(method, url, timeout=self.timeout, **kwargs)
        except Exception as e:
            if retry is None:
                raise

            retry = retry.increment(method=method, url=url)
            retry.sleep()
            response = self._request(method, endpoint, retry=retry, **kwargs)
        Api.__check_response_status_code(response)
        return response

    def post(self, endpoint, req_body=None, headers=None, retry=None):
        response = self._request(
            method="POST",
            endpoint=endpoint,
            data=req_body,
            headers=headers,
            stream=True,
            retry=retry,
        )
        return response

    def post_multipart(self, endpoint, parts=None, headers=None, retry=None):
        response = self._request(
            method="POST", endpoint=endpoint, files=parts, headers=headers, retry=retry
        )
        return response

    def put(self, endpoint, req_body=None, headers=None, retry=None):
        response = self._request(
            method="PUT", endpoint=endpoint, data=req_body, headers=headers, retry=retry
        )
        return response

    def patch(self, endpoint, req_body=None, headers=None, retry=None):
        response = self._request(
            method="PATCH",
            endpoint=endpoint,
            data=req_body,
            headers=headers,
            retry=retry,
        )
        return response

    def get(self, endpoint, params=None, headers=None, retry=None):
        response = self._request(
            method="GET", endpoint=endpoint, params=params, headers=headers, retry=retry
        )
        return response

    def head(self, endpoint, headers=None, retry=None):
        response = self._request(
            method="HEAD", endpoint=endpoint, headers=headers, retry=retry
        )
        return response

    def delete(self, endpoint, headers=None, retry=None):
        response = self._request(
            method="DELETE", endpoint=endpoint, headers=headers, retry=retry
        )
        return response



from __future__ import annotations
from google.protobuf.json_format import MessageToDict
from typing import List, Optional, Set, Tuple
from .proto import (
    PrimitiveType,
    PrivacySettings,
    ComputationConfiguration,
    TableSchema,
    ValidationConfiguration,
    Constraint,
    NamedColumn,
    ColumnType,
    SqlWorkerConfiguration,
    TableDependencyMapping,
)
from ...proto import (
    serialize_length_delimited,
    ComputeNodeFormat,
    parse_length_delimited,
)
from ..node import Node


class SqlCompute(Node):
    """
    Computation node to execute an SQL query.
    """

    def __init__(
        self,
        name: str,
        sql_statement: str,
        dependencies: List[Tuple[str, str]],
        *,
        privacy_settings: Optional[PrivacySettings] = None,
        constraints: Optional[List[Constraint]] = None,
    ) -> None:
        sql_worker_configuration = SqlWorkerConfiguration(
            computation=ComputationConfiguration(
                sqlStatement=sql_statement,
                privacySettings=privacy_settings,
                constraints=constraints,
                tableDependencyMappings=[
                    TableDependencyMapping(table=table, dependency=node_id)
                    for table, node_id in dependencies
                ],
            )
        )
        config = serialize_length_delimited(sql_worker_configuration)

        super().__init__(
            name,
            config=config,
            enclave_type="decentriq.sql-worker",
            dependencies=[node_id for _, node_id in dependencies],
            output_format=ComputeNodeFormat.ZIP,
        )


class SqlSchemaVerifier(Node):
    """
    Computation node to validate an input and provide the necessary types.
    """

    def __init__(
        self,
        name: str,
        input_data_node: str,
        columns,  # type: List[Tuple[str, PrimitiveType.V, bool]]
    ) -> None:
        named_columns = map(
            lambda c: NamedColumn(
                name=c[0], columnType=ColumnType(primitiveType=c[1], nullable=c[2])
            ),
            columns,
        )
        sql_worker_configuration = SqlWorkerConfiguration(
            validation=ValidationConfiguration(
                tableSchema=TableSchema(namedColumns=named_columns)
            )
        )
        config = serialize_length_delimited(sql_worker_configuration)

        super().__init__(
            name,
            config=config,
            enclave_type="decentriq.sql-worker",
            dependencies=[input_data_node],
            output_format=ComputeNodeFormat.ZIP,
        )


from ...decoder import SqlWorkerDecoder



import io

from typing import Optional

from .session import Session
from .storage import Key
from .keychain import Keychain


def provision_tabular_dataset_to_data_science_data_room(
    data: io.BytesIO,
    *,
    session: Session,
    key: Key,
    data_node: str,
    data_room_id: str,
    store_in_keychain: Optional[Keychain] = None,
    description: str = "",
) -> str:
    """
    Convenience function for uploading data to a tabular data node in a Data Science Data Room.

    **Parameters**:
    - `data`: The input data to be uploaded. Use one of the reader functions provided in this
        package to read CSV-like data.
    - `session`: The session with which to communicate with the enclave.
    - `key`: A key for encrypting the data to-be-uploaded.
    - `data_room_id`: To which data room the dataset should be published. This is the id you
        get when publishing a data room.
    - `store_in_keychain`: An optional keychain in which to store the dataset key.
    - `description`: An optional description of the dataset.

    **Returns**:
    The manifest hash (dataset id) in case the upload succeeded.
    """
    manifest_hash = session.client.upload_dataset(
        data,
        key,
        data_node,
        description=description,
        store_in_keychain=store_in_keychain,
    )
    session.publish_dataset(
        data_room_id, manifest_hash, leaf_id=f"{data_node}_leaf", key=key
    )
    return manifest_hash


def provision_raw_dataset_to_data_science_data_room(
    data: io.BytesIO,
    *,
    session: Session,
    key: Key,
    data_node: str,
    data_room_id: str,
    store_in_keychain: Optional[Keychain] = None,
    description: str = "",
) -> str:
    """
    Convenience function for uploading data to a raw leaf node in a Data Science Data Room.

    **Parameters**:
    - `data`: The input data to be uploaded. Use one of the reader functions provided in this
        package to read CSV-like data.
    - `session`: The session with which to communicate with the enclave.
    - `key`: A key for encrypting the data to-be-uploaded.
    - `data_room_id`: To which data room the dataset should be published. This is the id you
        get when publishing a data room.
    - `store_in_keychain`: An optional keychain in which to store the dataset key.
    - `description`: An optional description of the dataset.

    **Returns**:
    The manifest hash (dataset id) in case the upload succeeded.
    """
    manifest_hash = session.client.upload_dataset(
        data,
        key,
        data_node,
        description=description,
        store_in_keychain=store_in_keychain,
    )
    session.publish_dataset(
        data_room_id, manifest_hash, leaf_id=f"{data_node}_leaf", key=key
    )
    return manifest_hash


def provision_matching_dataset_to_lookalike_media_data_room(
    data: io.BytesIO,
    *,
    session: Session,
    key: Key,
    data_node: str,
    data_room_id: str,
    store_in_keychain: Optional[Keychain] = None,
    description: str = "",
) -> str:
    """
    Convenience function for uploading matching data to a Lookalike Media Data Room.

    **Parameters**:
    - `data`: The input data to be uploaded. Use one of the reader functions provided in this
        package to read CSV-like data.
    - `session`: The session with which to communicate with the enclave.
    - `key`: A key for encrypting the data to-be-uploaded.
    - `data_room_id`: To which data room the dataset should be published. This is the id you
        get when publishing a data room.
    - `store_in_keychain`: An optional keychain in which to store the dataset key.
    - `description`: An optional description of the dataset.

    **Returns**:
    The manifest hash (dataset id) in case the upload succeeded.
    """
    manifest_hash = session.client.upload_dataset(
        data,
        key,
        data_node,
        description=description,
        store_in_keychain=store_in_keychain,
    )
    session.publish_dataset(data_room_id, manifest_hash, leaf_id="matching", key=key)
    return manifest_hash



from __future__ import annotations

from google.protobuf.json_format import MessageToDict
from .proto import (
    parse_length_delimited,
)
from .proto import DriverTaskConfig
from .proto import PostWorkerConfiguration
from .proto import PermutiveWorkerConfiguration
from .proto import AzureBlobStorageWorkerConfiguration
from .proto import ContainerWorkerConfiguration
from .proto import DataSourceS3WorkerConfiguration
from .proto import DataSourceSnowflakeWorkerConfiguration
from .proto import DatasetSinkWorkerConfiguration
from .proto import GoogleAdManagerWorkerConfiguration
from .proto import GoogleDv360SinkWorkerConfiguration
from .proto import MetaSinkWorkerConfiguration
from .proto import PermutiveWorkerConfiguration
from .proto import S3SinkWorkerConfiguration
from .proto import SalesforceWorkerConfiguration
from .proto import SqlWorkerConfiguration


class GcgDriverDecoder:
    def decode(self, config: bytes):
        config_decoded = DriverTaskConfig()
        parse_length_delimited(config, config_decoded)
        config_decoded_json = MessageToDict(config_decoded)
        if config_decoded.HasField("staticContent"):
            content = config_decoded.staticContent.content.decode("utf-8")
            config_decoded_json["staticContent"]["content"] = content
        return config_decoded_json


class PostWorkerDecoder:
    def decode(self, config: bytes):
        config_decoded = PostWorkerConfiguration()
        parse_length_delimited(config, config_decoded)
        return MessageToDict(config_decoded)


class AzureBlobStorageWorkerDecoder:
    def decode(self, config: bytes):
        config_decoded = AzureBlobStorageWorkerConfiguration()
        parse_length_delimited(config, config_decoded)
        return MessageToDict(config_decoded)


class ContainerWorkerDecoder:
    def decode(self, config: bytes):
        config_decoded = ContainerWorkerConfiguration()
        parse_length_delimited(config, config_decoded)
        return MessageToDict(config_decoded)


class DataSourceS3WorkerDecoder:
    def decode(self, config: bytes):
        config_decoded = DataSourceS3WorkerConfiguration()
        parse_length_delimited(config, config_decoded)
        return MessageToDict(config_decoded)


class DataSourceSnowflakeWorkerDecoder:
    def decode(self, config: bytes):
        config_decoded = DataSourceSnowflakeWorkerConfiguration()
        parse_length_delimited(config, config_decoded)
        return MessageToDict(config_decoded)


class DatasetSinkWorkerDecoder:
    def decode(self, config: bytes):
        config_decoded = DatasetSinkWorkerConfiguration()
        parse_length_delimited(config, config_decoded)
        return MessageToDict(config_decoded)


class GoogleAdManagerWorkerDecoder:
    def decode(self, config: bytes):
        config_decoded = GoogleAdManagerWorkerConfiguration()
        parse_length_delimited(config, config_decoded)
        return MessageToDict(config_decoded)


class GoogleDv360SinkWorkerDecoder:
    def decode(self, config: bytes):
        config_decoded = GoogleDv360SinkWorkerConfiguration()
        parse_length_delimited(config, config_decoded)
        return MessageToDict(config_decoded)


class MetaSinkWorkerDecoder:
    def decode(self, config: bytes):
        config_decoded = MetaSinkWorkerConfiguration()
        parse_length_delimited(config, config_decoded)
        return MessageToDict(config_decoded)


class PermutiveWorkerDecoder:
    def decode(self, config: bytes):
        config_decoded = PermutiveWorkerConfiguration()
        parse_length_delimited(config, config_decoded)
        return MessageToDict(config_decoded)


class S3SinkWorkerDecoder:
    def decode(self, config: bytes):
        config_decoded = S3SinkWorkerConfiguration()
        parse_length_delimited(config, config_decoded)
        return MessageToDict(config_decoded)


class DataSourceSalesforceWorkerDecoder:
    def decode(self, config: bytes):
        config_decoded = SalesforceWorkerConfiguration()
        parse_length_delimited(config, config_decoded)
        return MessageToDict(config_decoded)


class SqlWorkerDecoder:
    def decode(self, config: bytes):
        config_decoded = SqlWorkerConfiguration()
        parse_length_delimited(config, config_decoded)
        return MessageToDict(config_decoded)
